@@ -1,4 +1,5 @@
 """Group handlers"""
+
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 import json
@@ -18,7 +19,7 @@ class _GroupAPIHandler(APIHandler):
             username = self.authenticator.normalize_username(username)
             user = self.find_user(username)
             if user is None:
-                raise web.HTTPError(400, "No such user: %s" % username)
+                raise web.HTTPError(400, f"No such user: {username}")
             users.append(user.orm_user)
         return users
 
@@ -86,7 +87,7 @@ class GroupListAPIHandler(_GroupAPIHandler):
         for name in groupnames:
             existing = orm.Group.find(self.db, name=name)
             if existing is not None:
-                raise web.HTTPError(409, "Group %s already exists" % name)
+                raise web.HTTPError(409, f"Group {name} already exists")
 
             usernames = model.get('users', [])
             # check that users exist
@@ -94,8 +95,9 @@ class GroupListAPIHandler(_GroupAPIHandler):
             # create the group
             self.log.info("Creating new group %s with %i users", name, len(users))
             self.log.debug("Users: %s", usernames)
-            group = orm.Group(name=name, users=users)
+            group = orm.Group(name=name)
             self.db.add(group)
+            group.users = users
             self.db.commit()
             created.append(group)
         self.write(json.dumps([self.group_model(group) for group in created]))
@@ -122,7 +124,7 @@ class GroupAPIHandler(_GroupAPIHandler):
 
         existing = orm.Group.find(self.db, name=group_name)
         if existing is not None:
-            raise web.HTTPError(409, "Group %s already exists" % group_name)
+            raise web.HTTPError(409, f"Group {group_name} already exists")
 
         usernames = model.get('users', [])
         # check that users exist
@@ -131,8 +133,9 @@ class GroupAPIHandler(_GroupAPIHandler):
         # create the group
         self.log.info("Creating new group %s with %i users", group_name, len(users))
         self.log.debug("Users: %s", usernames)
-        group = orm.Group(name=group_name, users=users)
+        group = orm.Group(name=group_name)
         self.db.add(group)
+        group.users = users
         self.db.commit()
         self.write(json.dumps(self.group_model(group)))
         self.set_status(201)
@@ -192,8 +195,25 @@ class GroupUsersAPIHandler(_GroupAPIHandler):
         self.write(json.dumps(self.group_model(group)))
 
 
+class GroupPropertiesAPIHandler(_GroupAPIHandler):
+    """Modify a group's properties"""
+
+    @needs_scope('groups')
+    def put(self, group_name):
+        group = self.find_group(group_name)
+        data = self.get_json_body()
+        # self._check_group_model(data)
+        if not isinstance(data, dict):
+            raise web.HTTPError(400, "Must specify properties")
+        self.log.info("Updating properties of group %s", group_name)
+        group.properties = data
+        self.db.commit()
+        self.write(json.dumps(self.group_model(group)))
+
+
 default_handlers = [
     (r"/api/groups", GroupListAPIHandler),
     (r"/api/groups/([^/]+)", GroupAPIHandler),
     (r"/api/groups/([^/]+)/users", GroupUsersAPIHandler),
+    (r"/api/groups/([^/]+)/properties", GroupPropertiesAPIHandler),
 ]
