@@ -1,4 +1,5 @@
 """Basic html-rendering handlers."""
+
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 import asyncio
@@ -11,10 +12,10 @@ from jinja2 import TemplateNotFound
 from tornado import web
 from tornado.httputil import url_concat
 
-from .. import __version__
+from .. import __version__, orm
 from ..metrics import SERVER_POLL_DURATION_SECONDS, ServerPollStatus
-from ..scopes import needs_scope
-from ..utils import maybe_future, url_escape_path, url_path_join
+from ..scopes import describe_raw_scopes, needs_scope
+from ..utils import maybe_future, url_escape_path, url_path_join, utcnow
 from .base import BaseHandler
 
 
@@ -72,7 +73,7 @@ class HomeHandler(BaseHandler):
             user=user,
             url=url,
             allow_named_servers=self.allow_named_servers,
-            named_server_limit_per_user=self.named_server_limit_per_user,
+            named_server_limit_per_user=await self.get_current_user_named_server_limit(),
             url_path_join=url_path_join,
             # can't use user.spawners because the stop method of User pops named servers from user.spawners when they're stopped
             spawners=user.orm_user._orm_spawners,
@@ -99,7 +100,9 @@ class SpawnHandler(BaseHandler):
             auth_state=auth_state,
             spawner_options_form=spawner_options_form,
             error_message=message,
-            url=self.request.uri,
+            url=url_concat(
+                self.request.uri, {"_xsrf": self.xsrf_token.decode('ascii')}
+            ),
             spawner=for_user.spawner,
         )
 
@@ -129,18 +132,18 @@ class SpawnHandler(BaseHandler):
         if server_name:
             if not self.allow_named_servers:
                 raise web.HTTPError(400, "Named servers are not enabled.")
-            if (
-                self.named_server_limit_per_user > 0
-                and server_name not in user.orm_spawners
-            ):
+
+            named_server_limit_per_user = (
+                await self.get_current_user_named_server_limit()
+            )
+
+            if named_server_limit_per_user > 0 and server_name not in user.orm_spawners:
                 named_spawners = list(user.all_spawners(include_default=False))
-                if self.named_server_limit_per_user <= len(named_spawners):
+                if named_server_limit_per_user <= len(named_spawners):
                     raise web.HTTPError(
                         400,
-                        "User {} already has the maximum of {} named servers."
-                        "  One must be deleted before a new server can be created".format(
-                            user.name, self.named_server_limit_per_user
-                        ),
+                        f"User {user.name} already has the maximum of {named_server_limit_per_user} named servers."
+                        "  One must be deleted before a new server can be created",
                     )
 
         if not self.allow_named_servers and user.running:
@@ -233,12 +236,12 @@ class SpawnHandler(BaseHandler):
         if for_user != user.name:
             user = self.find_user(for_user)
             if user is None:
-                raise web.HTTPError(404, "No such user: %s" % for_user)
+                raise web.HTTPError(404, f"No such user: {for_user}")
 
         spawner = user.get_spawner(server_name, replace_failed=True)
 
         if spawner.ready:
-            raise web.HTTPError(400, "%s is already running" % (spawner._log_name))
+            raise web.HTTPError(400, f"{spawner._log_name} is already running")
         elif spawner.pending:
             raise web.HTTPError(
                 400, f"{spawner._log_name} is pending {spawner.pending}"
@@ -248,7 +251,7 @@ class SpawnHandler(BaseHandler):
         for key, byte_list in self.request.body_arguments.items():
             form_options[key] = [bs.decode('utf8') for bs in byte_list]
         for key, byte_list in self.request.files.items():
-            form_options["%s_file" % key] = byte_list
+            form_options[f"{key}_file"] = byte_list
         try:
             self.log.debug(
                 "Triggering spawn with supplied form options for %s", spawner._log_name
@@ -342,7 +345,7 @@ class SpawnPendingHandler(BaseHandler):
         if for_user != current_user.name:
             user = self.find_user(for_user)
             if user is None:
-                raise web.HTTPError(404, "No such user: %s" % for_user)
+                raise web.HTTPError(404, f"No such user: {for_user}")
 
         if server_name and server_name not in user.spawners:
             raise web.HTTPError(404, f"{user.name} has no such server {server_name}")
@@ -402,7 +405,9 @@ class SpawnPendingHandler(BaseHandler):
                 page,
                 user=user,
                 spawner=spawner,
-                progress_url=spawner._progress_url,
+                progress_url=url_concat(
+                    spawner._progress_url, {"_xsrf": self.xsrf_token.decode('ascii')}
+                ),
                 auth_state=auth_state,
             )
             self.finish(html)
@@ -458,7 +463,7 @@ class AdminHandler(BaseHandler):
             auth_state=auth_state,
             admin_access=True,
             allow_named_servers=self.allow_named_servers,
-            named_server_limit_per_user=self.named_server_limit_per_user,
+            named_server_limit_per_user=await self.get_current_user_named_server_limit(),
             server_version=f'{__version__} {self.version_hash}',
             api_page_limit=self.settings["api_page_default_limit"],
             base_url=self.settings["base_url"],
@@ -478,7 +483,7 @@ class TokenPageHandler(BaseHandler):
         def sort_key(token):
             return (token.last_activity or never, token.created or never)
 
-        now = datetime.utcnow()
+        now = utcnow(with_tz=False)
 
         # group oauth client tokens by client id
         all_tokens = defaultdict(list)
@@ -546,6 +551,83 @@ class TokenPageHandler(BaseHandler):
         self.finish(html)
 
 
+class AcceptShareHandler(BaseHandler):
+    def _get_next_url(self, owner, spawner):
+        """Get next_url for a given owner/spawner"""
+        next_url = self.get_argument("next", "")
+        next_url = self._validate_next_url(next_url)
+        if next_url:
+            return next_url
+
+        # default behavior:
+        # if it's active, redirect to server URL
+        if spawner.name in owner.spawners:
+            spawner = owner.spawners[spawner.name]
+            if spawner.active:
+                # redirect to spawner url
+                next_url = owner.server_url(spawner.name)
+
+        if not next_url:
+            # spawner not active
+            # TODO: next_url not specified and not running, what do we do?
+            # for now, redirect as if it's running,
+            # but that's very likely to fail on "You can't launch this server"
+            # is there a better experience for this?
+            next_url = owner.server_url(spawner.name)
+        # validate again, which strips away host to just absolute path
+        return self._validate_next_url(next_url)
+
+    @web.authenticated
+    async def get(self):
+        code = self.get_argument("code")
+        share_code = orm.ShareCode.find(self.db, code=code)
+        if share_code is None:
+            raise web.HTTPError(404, "Share not found or expired")
+        if share_code.owner == self.current_user.orm_user:
+            raise web.HTTPError(403, "You can't share with yourself!")
+
+        scope_descriptions = describe_raw_scopes(
+            share_code.scopes,
+            username=self.current_user.name,
+        )
+        owner = self._user_from_orm(share_code.owner)
+        spawner = share_code.spawner
+        if spawner.name in owner.spawners:
+            spawner = owner.spawners[spawner.name]
+            spawner_ready = spawner.ready
+        else:
+            spawner_ready = False
+
+        html = await self.render_template(
+            'accept-share.html',
+            code=code,
+            owner=owner,
+            spawner=spawner,
+            spawner_ready=spawner_ready,
+            spawner_url=owner.server_url(spawner.name),
+            scope_descriptions=scope_descriptions,
+            next_url=self._get_next_url(owner, spawner),
+        )
+        self.finish(html)
+
+    @web.authenticated
+    def post(self):
+        code = self.get_argument("code")
+        self.log.debug("Looking up %s", code)
+        share_code = orm.ShareCode.find(self.db, code=code)
+        if share_code is None:
+            raise web.HTTPError(400, f"Invalid share code: {code}")
+        if share_code.owner == self.current_user.orm_user:
+            raise web.HTTPError(400, "You can't share with yourself!")
+        user = self.current_user
+        share = share_code.exchange(user.orm_user)
+        owner = self._user_from_orm(share.owner)
+        spawner = share.spawner
+
+        next_url = self._get_next_url(owner, spawner)
+        self.redirect(next_url)
+
+
 class ProxyErrorHandler(BaseHandler):
     """Handler for rendering proxy error pages"""
 
@@ -560,7 +642,7 @@ class ProxyErrorHandler(BaseHandler):
             message_html = ' '.join(
                 [
                     "Your server appears to be down.",
-                    "Try restarting it <a href='%s'>from the hub</a>" % hub_home,
+                    f"Try restarting it <a href='{hub_home}'>from the hub</a>",
                 ]
             )
         ns = dict(
@@ -573,9 +655,9 @@ class ProxyErrorHandler(BaseHandler):
         self.set_header('Content-Type', 'text/html')
         # render the template
         try:
-            html = await self.render_template('%s.html' % status_code, **ns)
+            html = await self.render_template(f'{status_code}.html', **ns)
         except TemplateNotFound:
-            self.log.debug("No template for %d", status_code)
+            self.log.debug("Using default error template for %d", status_code)
             html = await self.render_template('error.html', **ns)
 
         self.write(html)
@@ -602,6 +684,7 @@ default_handlers = [
     (r'/spawn/([^/]+)', SpawnHandler),
     (r'/spawn/([^/]+)/([^/]+)', SpawnHandler),
     (r'/token', TokenPageHandler),
+    (r'/accept-share', AcceptShareHandler),
     (r'/error/(\d+)', ProxyErrorHandler),
     (r'/health$', HealthCheckHandler),
     (r'/api/health$', HealthCheckHandler),
