@@ -1,13 +1,19 @@
 """sqlalchemy ORM tools for the state of the constellation of processes"""
+
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 import enum
 import json
+import numbers
+import secrets
 from base64 import decodebytes, encodebytes
-from datetime import datetime, timedelta
+from datetime import timedelta
+from functools import lru_cache, partial
+from itertools import chain
 
 import alembic.command
 import alembic.config
+import sqlalchemy
 from alembic.script import ScriptDirectory
 from sqlalchemy import (
     Boolean,
@@ -24,25 +30,26 @@ from sqlalchemy import (
     inspect,
     or_,
     select,
+    text,
 )
-from sqlalchemy.ext.declarative import declarative_base
 from sqlalchemy.orm import (
     Session,
-    backref,
+    declarative_base,
+    declared_attr,
     interfaces,
+    joinedload,
     object_session,
     relationship,
     sessionmaker,
 )
 from sqlalchemy.pool import StaticPool
-from sqlalchemy.sql.expression import bindparam
 from sqlalchemy.types import LargeBinary, Text, TypeDecorator
 from tornado.log import app_log
 
-from .utils import compare_token, hash_token, new_token, random_port
+from .utils import compare_token, hash_token, new_token, random_port, utcnow
 
 # top-level variable for easier mocking in tests
-utcnow = datetime.utcnow
+utcnow = partial(utcnow, with_tz=False)
 
 
 class JSONDict(TypeDecorator):
@@ -90,6 +97,10 @@ class JSONDict(TypeDecorator):
 class JSONList(JSONDict):
     """Represents an immutable structure as a json-encoded string (to be used for list type columns).
 
+    Accepts list, tuple, sets for assignment
+
+    Always read as a list
+
     Usage::
 
         JSONList(JSONDict)
@@ -97,8 +108,12 @@ class JSONList(JSONDict):
     """
 
     def process_bind_param(self, value, dialect):
-        if isinstance(value, list) and value is not None:
+        if isinstance(value, (list, tuple)):
             value = json.dumps(value)
+        if isinstance(value, set):
+            # serialize sets as ordered lists
+            value = json.dumps(sorted(value))
+
         return value
 
     def process_result_value(self, value, dialect):
@@ -138,6 +153,9 @@ class Server(Base):
     base_url = Column(Unicode(255), default='/')
     cookie_name = Column(Unicode(255), default='cookie')
 
+    service = relationship("Service", back_populates="server", uselist=False)
+    spawner = relationship("Spawner", back_populates="server", uselist=False)
+
     def __repr__(self):
         return f"<Server({self.ip}:{self.port})>"
 
@@ -145,19 +163,19 @@ class Server(Base):
 # lots of things have roles
 # mapping tables are the same for all of them
 
-_role_map_tables = []
+_role_associations = {}
 
-for has_role in (
+for entity in (
     'user',
     'group',
     'service',
 ):
-    role_map = Table(
-        f'{has_role}_role_map',
+    table = Table(
+        f'{entity}_role_map',
         Base.metadata,
         Column(
-            f'{has_role}_id',
-            ForeignKey(f'{has_role}s.id', ondelete='CASCADE'),
+            f'{entity}_id',
+            ForeignKey(f'{entity}s.id', ondelete='CASCADE'),
             primary_key=True,
         ),
         Column(
@@ -165,8 +183,12 @@ for has_role in (
             ForeignKey('roles.id', ondelete='CASCADE'),
             primary_key=True,
         ),
+        Column('managed_by_auth', Boolean, default=False, nullable=False),
+    )
+
+    _role_associations[entity] = type(
+        entity.title() + 'RoleMap', (Base,), {'__table__': table}
     )
-    _role_map_tables.append(role_map)
 
 
 class Role(Base):
@@ -177,17 +199,17 @@ class Role(Base):
     name = Column(Unicode(255), unique=True)
     description = Column(Unicode(1023))
     scopes = Column(JSONList, default=[])
-    users = relationship('User', secondary='user_role_map', backref='roles')
-    services = relationship('Service', secondary='service_role_map', backref='roles')
-    groups = relationship('Group', secondary='group_role_map', backref='roles')
+
+    users = relationship('User', secondary='user_role_map', back_populates='roles')
+    services = relationship(
+        'Service', secondary='service_role_map', back_populates='roles'
+    )
+    groups = relationship('Group', secondary='group_role_map', back_populates='roles')
+
+    managed_by_auth = Column(Boolean, default=False, nullable=False)
 
     def __repr__(self):
-        return "<{} {} ({}) - scopes: {}>".format(
-            self.__class__.__name__,
-            self.name,
-            self.description,
-            self.scopes,
-        )
+        return f"<{self.__class__.__name__} {self.name} ({self.description}) - scopes: {self.scopes}>"
 
     @classmethod
     def find(cls, db, name):
@@ -212,14 +234,24 @@ class Group(Base):
     __tablename__ = 'groups'
     id = Column(Integer, primary_key=True, autoincrement=True)
     name = Column(Unicode(255), unique=True)
-    users = relationship('User', secondary='user_group_map', backref='groups')
+    users = relationship('User', secondary='user_group_map', back_populates='groups')
+    properties = Column(JSONDict, default={})
+    roles = relationship(
+        'Role', secondary='group_role_map', back_populates='groups', lazy="selectin"
+    )
+
+    shared_with_me = relationship(
+        "Share",
+        back_populates="group",
+        cascade="all, delete-orphan",
+        lazy="selectin",
+    )
+
+    # used in some model fields to differentiate 'whoami'
+    kind = "group"
 
     def __repr__(self):
-        return "<%s %s (%i users)>" % (
-            self.__class__.__name__,
-            self.name,
-            len(self.users),
-        )
+        return f"<{self.__class__.__name__} {self.name}>"
 
     @classmethod
     def find(cls, db, name):
@@ -256,8 +288,15 @@ class User(Base):
     id = Column(Integer, primary_key=True, autoincrement=True)
     name = Column(Unicode(255), unique=True)
 
+    roles = relationship(
+        'Role',
+        secondary='user_role_map',
+        back_populates='users',
+        lazy="selectin",
+    )
+
     _orm_spawners = relationship(
-        "Spawner", backref="user", cascade="all, delete-orphan"
+        "Spawner", back_populates="user", cascade="all, delete-orphan"
     )
 
     @property
@@ -265,13 +304,57 @@ class User(Base):
         return {s.name: s for s in self._orm_spawners}
 
     admin = Column(Boolean(create_constraint=False), default=False)
-    created = Column(DateTime, default=datetime.utcnow)
+    created = Column(DateTime, default=utcnow)
     last_activity = Column(DateTime, nullable=True)
 
-    api_tokens = relationship("APIToken", backref="user", cascade="all, delete-orphan")
-    oauth_codes = relationship(
-        "OAuthCode", backref="user", cascade="all, delete-orphan"
+    api_tokens = relationship(
+        "APIToken", back_populates="user", cascade="all, delete-orphan"
     )
+    groups = relationship(
+        "Group",
+        secondary='user_group_map',
+        back_populates="users",
+        lazy="selectin",
+    )
+    oauth_codes = relationship(
+        "OAuthCode", back_populates="user", cascade="all, delete-orphan"
+    )
+
+    # sharing relationships
+    shares = relationship(
+        "Share",
+        back_populates="owner",
+        cascade="all, delete-orphan",
+        foreign_keys="Share.owner_id",
+    )
+    share_codes = relationship(
+        "ShareCode",
+        back_populates="owner",
+        cascade="all, delete-orphan",
+        foreign_keys="ShareCode.owner_id",
+    )
+    shared_with_me = relationship(
+        "Share",
+        back_populates="user",
+        cascade="all, delete-orphan",
+        foreign_keys="Share.user_id",
+        lazy="selectin",
+    )
+
+    @property
+    def all_shared_with_me(self):
+        """return all shares shared with me,
+
+        including via group
+        """
+
+        return list(
+            chain(
+                self.shared_with_me,
+                *[group.shared_with_me for group in self.groups],
+            )
+        )
+
     cookie_id = Column(Unicode(255), default=new_token, nullable=False, unique=True)
     # User.state is actually Spawner state
     # We will need to figure something else out if/when we have multiple spawners per user
@@ -280,13 +363,12 @@ class User(Base):
     # Encryption is handled elsewhere
     encrypted_auth_state = Column(LargeBinary)
 
+    # used in some model fields to differentiate whether an owner or actor
+    # is a user or service
+    kind = "user"
+
     def __repr__(self):
-        return "<{cls}({name} {running}/{total} running)>".format(
-            cls=self.__class__.__name__,
-            name=self.name,
-            total=len(self._orm_spawners),
-            running=sum(bool(s.server) for s in self._orm_spawners),
-        )
+        return f"<{self.__class__.__name__}({self.name} {sum(bool(s.server) for s in self._orm_spawners)}/{len(self._orm_spawners)} running)>"
 
     def new_api_token(self, token=None, **kwargs):
         """Create a new API token
@@ -310,15 +392,24 @@ class Spawner(Base):
 
     id = Column(Integer, primary_key=True, autoincrement=True)
     user_id = Column(Integer, ForeignKey('users.id', ondelete='CASCADE'))
+    user = relationship("User", back_populates="_orm_spawners")
 
     server_id = Column(Integer, ForeignKey('servers.id', ondelete='SET NULL'))
     server = relationship(
         Server,
-        backref=backref('spawner', uselist=False),
+        back_populates="spawner",
+        lazy="joined",
         single_parent=True,
         cascade="all, delete-orphan",
     )
 
+    shares = relationship(
+        "Share", back_populates="spawner", cascade="all, delete-orphan"
+    )
+    share_codes = relationship(
+        "ShareCode", back_populates="spawner", cascade="all, delete-orphan"
+    )
+
     state = Column(JSONDict)
     name = Column(Unicode(255))
 
@@ -336,7 +427,7 @@ class Spawner(Base):
     )
     oauth_client = relationship(
         'OAuthClient',
-        backref=backref("spawner", uselist=False),
+        back_populates="spawner",
         cascade="all, delete-orphan",
         single_parent=True,
     )
@@ -377,16 +468,39 @@ class Service(Base):
     # common user interface:
     name = Column(Unicode(255), unique=True)
     admin = Column(Boolean(create_constraint=False), default=False)
+    roles = relationship(
+        'Role', secondary='service_role_map', back_populates='services', lazy="selectin"
+    )
+
+    url = Column(Unicode(2047), nullable=True)
+
+    oauth_client_allowed_scopes = Column(JSONList, nullable=True)
+
+    info = Column(JSONDict, nullable=True)
+
+    display = Column(Boolean, nullable=True)
+
+    oauth_no_confirm = Column(Boolean, nullable=True)
+
+    command = Column(JSONList, nullable=True)
+
+    cwd = Column(Unicode(4095), nullable=True)
+
+    environment = Column(JSONDict, nullable=True)
+
+    user = Column(Unicode(255), nullable=True)
+
+    from_config = Column(Boolean, default=True)
 
     api_tokens = relationship(
-        "APIToken", backref="service", cascade="all, delete-orphan"
+        "APIToken", back_populates="service", cascade="all, delete-orphan"
     )
 
     # service-specific interface
     _server_id = Column(Integer, ForeignKey('servers.id', ondelete='SET NULL'))
     server = relationship(
         Server,
-        backref=backref('service', uselist=False),
+        back_populates="service",
         single_parent=True,
         cascade="all, delete-orphan",
     )
@@ -400,13 +514,17 @@ class Service(Base):
             ondelete='SET NULL',
         ),
     )
+
     oauth_client = relationship(
         'OAuthClient',
-        backref=backref("service", uselist=False),
+        back_populates="service",
         cascade="all, delete-orphan",
         single_parent=True,
     )
 
+    # used in some model fields to differentiate 'whoami'
+    kind = "service"
+
     def new_api_token(self, token=None, **kwargs):
         """Create a new API token
         If `token` is given, load that token.
@@ -446,6 +564,14 @@ class Expiring:
         else:
             return None
 
+    @property
+    def expired(self):
+        """Is this object expired?"""
+        if not self.expires_at:
+            return False
+        else:
+            return self.expires_in <= 0
+
     @classmethod
     def purge_expired(cls, db):
         """Purge expired API Tokens from the database"""
@@ -478,7 +604,7 @@ class Hashed(Expiring):
 
     @property
     def token(self):
-        raise AttributeError("token is write-only")
+        raise AttributeError(f"{self.__class__.__name__}.token is write-only")
 
     @token.setter
     def token(self, token):
@@ -506,12 +632,13 @@ class Hashed(Expiring):
         """Check if a token is acceptable"""
         if len(token) < cls.min_length:
             raise ValueError(
-                "Tokens must be at least %i characters, got %r"
-                % (cls.min_length, token)
+                f"{cls.__name__}.token must be at least {cls.min_length} characters, got {len(token)}: {token[: cls.prefix_length]}..."
             )
         found = cls.find(db, token)
         if found:
-            raise ValueError("Collision on token: %s..." % token[: cls.prefix_length])
+            raise ValueError(
+                f"Collision on {cls.__name__}: {token[: cls.prefix_length]}..."
+            )
 
     @classmethod
     def find_prefix(cls, db, token):
@@ -541,12 +668,348 @@ class Hashed(Expiring):
         `kind='user'` only returns API tokens for users
         `kind='service'` only returns API tokens for services
         """
-        prefix_match = cls.find_prefix(db, token)
+        prefix_match = cls.find_prefix(db, token).options(
+            joinedload(cls.user), joinedload(cls.service)
+        )
+
         for orm_token in prefix_match:
             if orm_token.match(token):
                 return orm_token
 
 
+class _Share:
+    """Common columns for Share and ShareCode"""
+
+    id = Column(Integer, primary_key=True, autoincrement=True)
+    created_at = Column(DateTime, nullable=False, default=utcnow)
+
+    # TODO: owner_id and spawner_id columns don't need `@declared_attr` when we can require sqlalchemy 2
+
+    # the owner of the shared server
+    # this is redundant with spawner.user, but saves a join
+    @declared_attr
+    def owner_id(self):
+        return Column(Integer, ForeignKey('users.id', ondelete="CASCADE"))
+
+    @declared_attr
+    def owner(self):
+        # table name happens to be appropriate 'shares', 'share_codes'
+        # could be another, more explicit attribute, but the values would be the same
+        return relationship(
+            "User",
+            back_populates=self.__tablename__,
+            foreign_keys=[self.owner_id],
+            lazy="selectin",
+        )
+
+    # the spawner the share is for
+    @declared_attr
+    def spawner_id(self):
+        return Column(Integer, ForeignKey('spawners.id', ondelete="CASCADE"))
+
+    @declared_attr
+    def spawner(self):
+        return relationship(
+            "Spawner",
+            back_populates=self.__tablename__,
+            lazy="selectin",
+        )
+
+    # the permissions granted (!server filter will always be applied)
+    scopes = Column(JSONList)
+    expires_at = Column(DateTime, nullable=True)
+
+    @classmethod
+    def apply_filter(cls, scopes, spawner):
+        """Apply our filter, ensures all scopes have appropriate !server filter
+
+        Any other filters will raise ValueError.
+        """
+        return cls._apply_filter(frozenset(scopes), spawner.user.name, spawner.name)
+
+    @staticmethod
+    @lru_cache
+    def _apply_filter(scopes, owner_name, server_name):
+        """
+        implementation of Share.apply_filter
+
+        Static method so @lru_cache is persisted across instances
+        """
+        filtered_scopes = []
+        server_filter = f"server={owner_name}/{server_name}"
+        for scope in scopes:
+            base_scope, _, filter = scope.partition("!")
+            if filter and filter != server_filter:
+                raise ValueError(
+                    f"!{filter} not allowed on sharing {scope}, only !{server_filter}"
+                )
+            filtered_scopes.append(f"{base_scope}!{server_filter}")
+        return frozenset(filtered_scopes)
+
+
+class Share(_Share, Expiring, Base):
+    """A single record of a sharing permission
+
+    granted by one user to another user (or group)
+
+    Restricted to a single server.
+    """
+
+    __tablename__ = "shares"
+
+    # who the share is granted to (user or group)
+    user_id = Column(Integer, ForeignKey('users.id', ondelete="CASCADE"), nullable=True)
+    user = relationship(
+        "User", back_populates="shared_with_me", foreign_keys=[user_id], lazy="selectin"
+    )
+
+    group_id = Column(
+        Integer, ForeignKey('groups.id', ondelete="CASCADE"), nullable=True
+    )
+    group = relationship("Group", back_populates="shared_with_me", lazy="selectin")
+
+    def __repr__(self):
+        if self.user:
+            kind = "user"
+            name = self.user.name
+        elif self.group:
+            kind = "group"
+            name = self.group.name
+        else:  # pragma: no cover
+            kind = "deleted"
+            name = "unknown"
+
+        if self.owner and self.spawner:
+            server_name = f"{self.owner.name}/{self.spawner.name}"
+        else:  # pragma: n cover
+            server_name = "unknown/deleted"
+
+        return f"<{self.__class__.__name__}(server={server_name}, scopes={self.scopes}, {kind}={name})>"
+
+    @staticmethod
+    def _share_with_key(share_with):
+        """Get the field name for share with
+
+        either group_id or user_id, depending on type of share_with
+
+        raises TypeError if neither User nor Group
+        """
+        if isinstance(share_with, User):
+            return "user_id"
+        elif isinstance(share_with, Group):
+            return "group_id"
+        else:
+            raise TypeError(
+                f"Can only share with orm.User or orm.Group, not {share_with!r}"
+            )
+
+    @classmethod
+    def find(cls, db, spawner, share_with):
+        """Find an existing
+
+        Shares are unique for a given (spawner, user)
+        """
+
+        filter_by = {
+            cls._share_with_key(share_with): share_with.id,
+            "spawner_id": spawner.id,
+            "owner_id": spawner.user.id,
+        }
+        return db.query(Share).filter_by(**filter_by).one_or_none()
+
+    @staticmethod
+    def _get_log_name(spawner, share_with):
+        """construct log snippet to refer to the share"""
+        return (
+            f"{share_with.kind}:{share_with.name} on {spawner.user.name}/{spawner.name}"
+        )
+
+    @property
+    def _log_name(self):
+        return self._get_log_name(self.spawner, self.user or self.group)
+
+    @classmethod
+    def grant(cls, db, spawner, share_with, scopes=None):
+        """Grant shared permissions for a server
+
+        Updates existing Share if there is one,
+        otherwise creates a new Share
+        """
+        if scopes is None:
+            scopes = frozenset(
+                [f"access:servers!server={spawner.user.name}/{spawner.name}"]
+            )
+        scopes = cls._apply_filter(frozenset(scopes), spawner.user.name, spawner.name)
+
+        if not scopes:
+            raise ValueError("Must specify scopes to grant.")
+
+        # 1. lookup existing share and update
+        share = cls.find(db, spawner, share_with)
+        share_with_log = cls._get_log_name(spawner, share_with)
+        if share is not None:
+            # update existing permissions in-place
+            # extend permissions
+            existing_scopes = set(share.scopes)
+            added_scopes = set(scopes).difference(existing_scopes)
+            if not added_scopes:
+                app_log.info(f"No new scopes for {share_with_log}")
+                return share
+            new_scopes = sorted(existing_scopes | added_scopes)
+            app_log.info(f"Granting scopes {sorted(added_scopes)} for {share_with_log}")
+            share.scopes = new_scopes
+            db.commit()
+        else:
+            # no share for (spawner, share_with), create a new one
+            app_log.info(f"Sharing scopes {sorted(scopes)} for {share_with_log}")
+            share = cls(
+                created_at=cls.now(),
+                # copy shared fields
+                owner=spawner.user,
+                spawner=spawner,
+                scopes=sorted(scopes),
+            )
+            if share_with.kind == "user":
+                share.user = share_with
+            elif share_with.kind == "group":
+                share.group = share_with
+            else:
+                raise TypeError(f"Expected user or group, got {share_with!r}")
+            db.add(share)
+            db.commit()
+        return share
+
+    @classmethod
+    def revoke(cls, db, spawner, share_with, scopes=None):
+        """Revoke permissions for share_with on `spawner`
+
+        If scopes are not specified, all scopes are revoked
+        """
+        share = cls.find(db, spawner, share_with)
+        if share is None:
+            _log_name = cls._get_log_name(spawner, share_with)
+            app_log.info(f"No permissions to revoke from {_log_name}")
+            return
+        else:
+            _log_name = share._log_name
+
+        if scopes is None:
+            app_log.info(f"Revoked all permissions from {_log_name}")
+            db.delete(share)
+            db.commit()
+            return None
+
+        # update scopes
+        new_scopes = [scope for scope in share.scopes if scope not in scopes]
+        revoked_scopes = [scope for scope in scopes if scope in set(share.scopes)]
+        if new_scopes == share.scopes:
+            app_log.info(f"No change in scopes for {_log_name}")
+            return share
+        elif not new_scopes:
+            # revoked all scopes, delete the Share
+            app_log.info(f"Revoked all permissions from {_log_name}")
+            db.delete(share)
+            db.commit()
+        else:
+            app_log.info(f"Revoked {revoked_scopes} from {_log_name}")
+            share.scopes = new_scopes
+            db.commit()
+
+        if new_scopes:
+            return share
+        else:
+            return None
+
+
+class ShareCode(_Share, Hashed, Base):
+    """A code that can be exchanged for a Share
+
+    Ultimately, the same as a Share, but has a 'code'
+    instead of a user or group that it is shared with.
+    The code can be exchanged to create or update an actual Share.
+    """
+
+    __tablename__ = "share_codes"
+
+    hashed = Column(Unicode(255), unique=True)
+    prefix = Column(Unicode(16), index=True)
+    exchange_count = Column(Integer, default=0)
+    last_exchanged_at = Column(DateTime, nullable=True, default=None)
+
+    _code_bytes = 32
+    default_expires_in = 86400
+
+    def __repr__(self):
+        if self.owner and self.spawner:
+            server_name = f"{self.owner.name}/{self.spawner.name}"
+        else:
+            server_name = "unknown/deleted"
+
+        return f"<{self.__class__.__name__}(server={server_name}, scopes={self.scopes}, expires_at={self.expires_at})>"
+
+    @classmethod
+    def new(
+        cls,
+        db,
+        spawner,
+        *,
+        scopes,
+        expires_in=None,
+        **kwargs,
+    ):
+        """Create a new ShareCode"""
+        app_log.info(f"Creating share code for {spawner.user.name}/{spawner.name}")
+        # verify scopes have the necessary filter
+        kwargs["scopes"] = sorted(cls.apply_filter(scopes, spawner))
+        if not expires_in:
+            expires_in = cls.default_expires_in
+        kwargs["expires_at"] = utcnow() + timedelta(seconds=expires_in)
+        kwargs["spawner"] = spawner
+        kwargs["owner"] = spawner.user
+        code = secrets.token_urlsafe(cls._code_bytes)
+
+        # create the ShareCode
+        share_code = cls(**kwargs)
+        # setting Hashed.token property sets the `hashed` column in the db
+        share_code.token = code
+        # actually put it in the db
+        db.add(share_code)
+        db.commit()
+        return (share_code, code)
+
+    @classmethod
+    def find(cls, db, code, *, spawner=None):
+        """Lookup a single ShareCode by code"""
+        prefix_match = cls.find_prefix(db, code)
+        if spawner:
+            prefix_match = prefix_match.filter_by(spawner_id=spawner.id)
+        for share_code in prefix_match:
+            if share_code.match(code):
+                return share_code
+
+    def exchange(self, share_with):
+        """exchange a ShareCode for a Share
+
+        share_with can be a User or a Group.
+        """
+        db = inspect(self).session
+        share_code_log = f"Share code {self.prefix}..."
+        if self.expired:
+            db.delete(self)
+            db.commit()
+            raise ValueError(f"{share_code_log} expired")
+
+        share_with_log = f"{share_with.kind}:{share_with.name} on {self.owner.name}/{self.spawner.name}"
+        app_log.info(f"Exchanging {share_code_log} for {share_with_log}")
+        share = Share.grant(db, self.spawner, share_with, self.scopes)
+        # note: we count exchanges, even if they don't modify the permissions
+        # (e.g. one user exchanging the same code twice)
+        self.exchange_count += 1
+        self.last_exchanged_at = self.now()
+        db.commit()
+        return share
+
+
 # ------------------------------------
 # OAuth tables
 # ------------------------------------
@@ -577,6 +1040,10 @@ class APIToken(Hashed, Base):
         nullable=True,
     )
 
+    user = relationship("User", back_populates="api_tokens")
+    service = relationship("Service", back_populates="api_tokens")
+    oauth_client = relationship("OAuthClient", back_populates="access_tokens")
+
     id = Column(Integer, primary_key=True)
     hashed = Column(Unicode(255), unique=True)
     prefix = Column(Unicode(16), index=True)
@@ -611,8 +1078,8 @@ class APIToken(Hashed, Base):
     session_id = Column(Unicode(255), nullable=True)
 
     # token metadata for bookkeeping
-    now = datetime.utcnow  # for expiry
-    created = Column(DateTime, default=datetime.utcnow)
+    now = utcnow  # for expiry
+    created = Column(DateTime, default=utcnow)
     expires_at = Column(DateTime, default=None, nullable=True)
     last_activity = Column(DateTime)
     note = Column(Unicode(1023))
@@ -629,13 +1096,7 @@ class APIToken(Hashed, Base):
             # this shouldn't happen
             kind = 'owner'
             name = 'unknown'
-        return "<{cls}('{pre}...', {kind}='{name}', client_id={client_id!r})>".format(
-            cls=self.__class__.__name__,
-            pre=self.prefix,
-            kind=kind,
-            name=name,
-            client_id=self.client_id,
-        )
+        return f"<{self.__class__.__name__}('{self.prefix}...', {kind}='{name}', client_id={self.client_id!r})>"
 
     @classmethod
     def find(cls, db, token, *, kind=None):
@@ -652,7 +1113,7 @@ class APIToken(Hashed, Base):
         elif kind == 'service':
             prefix_match = prefix_match.filter(cls.service_id != None)
         elif kind is not None:
-            raise ValueError("kind must be 'user', 'service', or None, not %r" % kind)
+            raise ValueError(f"kind must be 'user', 'service', or None, not {kind!r}")
         for orm_token in prefix_match:
             if orm_token.match(token):
                 if not orm_token.client_id:
@@ -750,6 +1211,7 @@ class APIToken(Hashed, Base):
             session_id=session_id,
             scopes=list(scopes),
         )
+        db.add(orm_token)
         orm_token.token = token
         if user:
             assert user.id is not None
@@ -757,10 +1219,20 @@ class APIToken(Hashed, Base):
         else:
             assert service.id is not None
             orm_token.service = service
-        if expires_in is not None:
+        if expires_in:
+            if not isinstance(expires_in, numbers.Real):
+                raise TypeError(
+                    f"expires_in must be a positive integer or null, not {expires_in!r}"
+                )
+            expires_in = int(expires_in)
+            # tokens must always expire in the future
+            if expires_in < 1:
+                raise ValueError(
+                    f"expires_in must be a positive integer or null, not {expires_in!r}"
+                )
+
             orm_token.expires_at = cls.now() + timedelta(seconds=expires_in)
 
-        db.add(orm_token)
         db.commit()
         return token
 
@@ -782,18 +1254,26 @@ class OAuthCode(Expiring, Base):
     client_id = Column(
         Unicode(255), ForeignKey('oauth_clients.identifier', ondelete='CASCADE')
     )
+    client = relationship(
+        "OAuthClient",
+        back_populates="codes",
+    )
     code = Column(Unicode(36))
     expires_at = Column(Integer)
     redirect_uri = Column(Unicode(1023))
     session_id = Column(Unicode(255))
     # state = Column(Unicode(1023))
     user_id = Column(Integer, ForeignKey('users.id', ondelete='CASCADE'))
+    user = relationship(
+        "User",
+        back_populates="oauth_codes",
+    )
 
     scopes = Column(JSONList, default=[])
 
     @staticmethod
     def now():
-        return datetime.utcnow().timestamp()
+        return utcnow(with_tz=True).timestamp()
 
     @classmethod
     def find(cls, db, code):
@@ -801,6 +1281,10 @@ class OAuthCode(Expiring, Base):
             db.query(cls)
             .filter(cls.code == code)
             .filter(or_(cls.expires_at == None, cls.expires_at >= cls.now()))
+            .options(
+                # load user with the code
+                joinedload(cls.user, innerjoin=True),
+            )
             .first()
         )
 
@@ -822,10 +1306,22 @@ class OAuthClient(Base):
     def client_id(self):
         return self.identifier
 
+    spawner = relationship(
+        "Spawner",
+        back_populates="oauth_client",
+        uselist=False,
+    )
+    service = relationship(
+        "Service",
+        back_populates="oauth_client",
+        uselist=False,
+    )
     access_tokens = relationship(
-        APIToken, backref='oauth_client', cascade='all, delete-orphan'
+        APIToken, back_populates='oauth_client', cascade='all, delete-orphan'
+    )
+    codes = relationship(
+        OAuthCode, back_populates='client', cascade='all, delete-orphan'
     )
-    codes = relationship(OAuthCode, backref='client', cascade='all, delete-orphan')
 
     # these are the scopes an oauth client is allowed to request
     # *not* the scopes of the client itself
@@ -901,23 +1397,23 @@ def register_ping_connection(engine):
     https://docs.sqlalchemy.org/en/rel_1_1/core/pooling.html#disconnect-handling-pessimistic
     """
 
-    @event.listens_for(engine, "engine_connect")
-    def ping_connection(connection, branch):
-        if branch:
-            # "branch" refers to a sub-connection of a connection,
-            # we don't want to bother pinging on these.
-            return
-
+    # listeners are normally registered as a decorator,
+    # but we need two different signatures to avoid SAWarning:
+    #    The argument signature for the "ConnectionEvents.engine_connect" event listener has changed
+    # while we support sqla 1.4 and 2.0.
+    # @event.listens_for(engine, "engine_connect")
+    def ping_connection(connection):
         # turn off "close with result".  This flag is only used with
         # "connectionless" execution, otherwise will be False in any case
         save_should_close_with_result = connection.should_close_with_result
         connection.should_close_with_result = False
 
         try:
-            # run a SELECT 1.   use a core select() so that
+            # run a SELECT 1. use a core select() so that
             # the SELECT of a scalar value without a table is
             # appropriately formatted for the backend
-            connection.scalar(select([1]))
+            with connection.begin() as transaction:
+                connection.scalar(select(1))
         except exc.DBAPIError as err:
             # catch SQLAlchemy's DBAPIError, which is a wrapper
             # for the DBAPI's exception.  It includes a .connection_invalidated
@@ -932,13 +1428,25 @@ def register_ping_connection(engine):
                 # itself and establish a new connection.  The disconnect detection
                 # here also causes the whole connection pool to be invalidated
                 # so that all stale connections are discarded.
-                connection.scalar(select([1]))
+                with connection.begin() as transaction:
+                    connection.scalar(select(1))
             else:
                 raise
         finally:
             # restore "close with result"
             connection.should_close_with_result = save_should_close_with_result
 
+    # sqla v1/v2 compatible invocation of @event.listens_for:
+    def ping_connection_v1(connection, branch=None):
+        """sqlalchemy < 2.0 compatibility"""
+        return ping_connection(connection)
+
+    if int(sqlalchemy.__version__.split(".", 1)[0]) >= 2:
+        listener = ping_connection
+    else:
+        listener = ping_connection_v1
+    event.listens_for(engine, "engine_connect")(listener)
+
 
 def check_db_revision(engine):
     """Check the JupyterHub database revision
@@ -956,7 +1464,10 @@ def check_db_revision(engine):
 
     from .dbutil import _temp_alembic_ini
 
-    with _temp_alembic_ini(engine.url) as ini:
+    # alembic needs the password if it's in the URL
+    engine_url = engine.url.render_as_string(hide_password=False)
+
+    with _temp_alembic_ini(engine_url) as ini:
         cfg = alembic.config.Config(ini)
         scripts = ScriptDirectory.from_config(cfg)
         head = scripts.get_heads()[0]
@@ -991,19 +1502,17 @@ def check_db_revision(engine):
 
     # check database schema version
     # it should always be defined at this point
-    alembic_revision = engine.execute(
-        'SELECT version_num FROM alembic_version'
-    ).first()[0]
+    with engine.begin() as connection:
+        alembic_revision = connection.execute(
+            text('SELECT version_num FROM alembic_version')
+        ).first()[0]
     if alembic_revision == head:
         app_log.debug("database schema version found: %s", alembic_revision)
-        pass
     else:
         raise DatabaseSchemaMismatch(
-            "Found database schema version {found} != {head}. "
+            f"Found database schema version {alembic_revision} != {head}. "
             "Backup your database and run `jupyterhub upgrade-db`"
-            " to upgrade to the latest schema.".format(
-                found=alembic_revision, head=head
-            )
+            " to upgrade to the latest schema."
         )
 
 
@@ -1011,13 +1520,16 @@ def mysql_large_prefix_check(engine):
     """Check mysql has innodb_large_prefix set"""
     if not str(engine.url).startswith('mysql'):
         return False
-    variables = dict(
-        engine.execute(
-            'show variables where variable_name like '
-            '"innodb_large_prefix" or '
-            'variable_name like "innodb_file_format";'
-        ).fetchall()
-    )
+    with engine.begin() as connection:
+        variables = dict(
+            connection.execute(
+                text(
+                    'show variables where variable_name like '
+                    '"innodb_large_prefix" or '
+                    'variable_name like "innodb_file_format";'
+                )
+            ).fetchall()
+        )
     if (
         variables.get('innodb_file_format', 'Barracuda') == 'Barracuda'
         and variables.get('innodb_large_prefix', 'ON') == 'ON'
@@ -1042,6 +1554,8 @@ def new_session_factory(
     elif url.startswith('mysql'):
         kwargs.setdefault('pool_recycle', 60)
 
+    kwargs.setdefault("future", True)
+
     if url.endswith(':memory:'):
         # If we're using an in-memory database, ensure that only one connection
         # is ever created.
