@@ -23,27 +23,31 @@ Fixtures to add functionality or spawning behavior
 - `slow_bad_spawn`
 
 """
+
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 import asyncio
 import copy
-import inspect
 import os
 import sys
-from functools import partial
-from getpass import getuser
 from subprocess import TimeoutExpired
 from unittest import mock
 
 from pytest import fixture, raises
-from tornado import ioloop
+from sqlalchemy import event
 from tornado.httpclient import HTTPError
 from tornado.platform.asyncio import AsyncIOMainLoop
 
 import jupyterhub.services.service
 
 from .. import crypto, orm, scopes
-from ..roles import create_role, get_default_roles, mock_roles, update_roles
+from ..roles import (
+    assign_default_roles,
+    create_role,
+    get_default_roles,
+    mock_roles,
+    update_roles,
+)
 from ..utils import random_port
 from . import mocking
 from .mocking import MockHub
@@ -60,7 +64,7 @@ def ssl_tmpdir(tmpdir_factory):
 
 
 @fixture(scope='module')
-def app(request, io_loop, ssl_tmpdir):
+async def app(request, io_loop, ssl_tmpdir):
     """Mock a jupyterhub app for testing"""
     mocked_app = None
     ssl_enabled = getattr(
@@ -72,10 +76,6 @@ def app(request, io_loop, ssl_tmpdir):
 
     mocked_app = MockHub.instance(**kwargs)
 
-    async def make_app():
-        await mocked_app.initialize([])
-        await mocked_app.start()
-
     def fin():
         # disconnect logging during cleanup because pytest closes captured FDs prematurely
         mocked_app.log.handlers = []
@@ -83,10 +83,11 @@ def app(request, io_loop, ssl_tmpdir):
         try:
             mocked_app.stop()
         except Exception as e:
-            print("Error stopping Hub: %s" % e, file=sys.stderr)
+            print(f"Error stopping Hub: {e}", file=sys.stderr)
 
     request.addfinalizer(fin)
-    io_loop.run_sync(make_app)
+    await mocked_app.initialize([])
+    await mocked_app.start()
     return mocked_app
 
 
@@ -108,18 +109,18 @@ def auth_state_enabled(app):
 @fixture
 def db():
     """Get a db session"""
-    global _db
-    if _db is None:
-        # make sure some initial db contents are filled out
-        # specifically, the 'default' jupyterhub oauth client
-        app = MockHub(db_url='sqlite:///:memory:')
-        app.init_db()
-        _db = app.db
-        for role in get_default_roles():
-            create_role(_db, role)
-        user = orm.User(name=getuser())
-        _db.add(user)
-        _db.commit()
+    # make sure some initial db contents are filled out
+    # specifically, the 'default' jupyterhub oauth client
+    app = MockHub(db_url='sqlite:///:memory:')
+    app.init_db()
+    _db = app.db
+    for role in get_default_roles():
+        create_role(_db, role)
+    user = orm.User(name="user")
+    _db.add(user)
+    _db.commit()
+    assign_default_roles(_db, user)
+    _db.commit()
     return _db
 
 
@@ -133,7 +134,12 @@ def event_loop(request):
 
 @fixture(scope='module')
 async def io_loop(event_loop, request):
-    """Same as pytest-tornado.io_loop, but re-scoped to module-level"""
+    """Mostly obsolete fixture for tornado event loop
+
+    Main purpose is to register cleanup (close) after we're done with the loop.
+    The main reason to depend on this fixture is to ensure your cleanup
+    happens before the io_loop is closed.
+    """
     io_loop = AsyncIOMainLoop()
     assert asyncio.get_event_loop() is event_loop
     assert io_loop.asyncio_loop is event_loop
@@ -146,11 +152,13 @@ async def io_loop(event_loop, request):
 
 
 @fixture(autouse=True)
-def cleanup_after(request, io_loop):
+async def cleanup_after(request, io_loop):
     """function-scoped fixture to shutdown user servers
 
     allows cleanup of servers between tests
     without having to launch a whole new app
+
+    depends on io_loop to ensure it runs before things are closed.
     """
 
     try:
@@ -165,19 +173,39 @@ def cleanup_after(request, io_loop):
         app = MockHub.instance()
         if app.db_file.closed:
             return
-        for uid, user in list(app.users.items()):
+
+        # cleanup users
+        for orm_user in app.db.query(orm.User):
+            user = app.users[orm_user]
             for name, spawner in list(user.spawners.items()):
                 if spawner.active:
                     try:
-                        io_loop.run_sync(lambda: app.proxy.delete_user(user, name))
+                        await app.proxy.delete_user(user, name)
                     except HTTPError:
                         pass
-                    io_loop.run_sync(lambda: user.stop(name))
+                    print(f"Stopping leftover server {spawner._log_name}")
+                    await user.stop(name)
             if user.name not in {'admin', 'user'}:
-                app.users.delete(uid)
+                app.log.debug(f"Deleting test user {user.name}")
+                app.users.delete(user.id)
         # delete groups
         for group in app.db.query(orm.Group):
+            app.log.debug(f"Deleting test group {group.name}")
             app.db.delete(group)
+        # delete shares
+        for share in app.db.query(orm.Share):
+            app.log.debug(f"Deleting test share {share}")
+            app.db.delete(share)
+
+        # clear services
+        for name, service in app._service_map.items():
+            if service.managed:
+                service.stop()
+        for orm_service in app.db.query(orm.Service):
+            if orm_service.oauth_client:
+                app.oauth_provider.remove_client(orm_service.oauth_client_id)
+            app.db.delete(orm_service)
+        app._service_map.clear()
         app.db.commit()
 
 
@@ -218,6 +246,7 @@ def admin_user(app, username):
 
 
 _groupname_counter = 0
+_rolename_counter = 0
 
 
 def new_group_name(prefix='testgroup'):
@@ -227,6 +256,13 @@ def new_group_name(prefix='testgroup'):
     return f'{prefix}-{_groupname_counter}'
 
 
+def new_role_name(prefix='testrole'):
+    """Return a new unique role name"""
+    global _rolename_counter
+    _rolename_counter += 1
+    return f'{prefix}-{_rolename_counter}'
+
+
 @fixture
 def groupname():
     """allocate a temporary group name
@@ -250,6 +286,22 @@ def group(app):
     yield group
 
 
+@fixture
+def role(app):
+    """Fixture for creating a temporary role
+
+    Each time the fixture is used, a new role is created
+
+    The role is deleted after the test
+    """
+    role = orm.Role(name=new_role_name())
+    app.db.add(role)
+    app.db.commit()
+    yield role
+    app.db.delete(role)
+    app.db.commit()
+
+
 class MockServiceSpawner(jupyterhub.services.service._ServiceSpawner):
     """mock services for testing.
 
@@ -259,10 +311,7 @@ class MockServiceSpawner(jupyterhub.services.service._ServiceSpawner):
     poll_interval = 1
 
 
-_mock_service_counter = 0
-
-
-def _mockservice(request, app, external=False, url=False):
+async def _mockservice(request, app, name, external=False, url=False):
     """
     Add a service to the application
 
@@ -278,9 +327,6 @@ def _mockservice(request, app, external=False, url=False):
           If True, register the service at a URL
           (as opposed to headless, API-only).
     """
-    global _mock_service_counter
-    _mock_service_counter += 1
-    name = 'mock-service-%i' % _mock_service_counter
     spec = {'name': name, 'command': mockservice_cmd, 'admin': True}
     if url:
         if app.internal_ssl:
@@ -289,10 +335,9 @@ def _mockservice(request, app, external=False, url=False):
             spec['url'] = 'http://127.0.0.1:%i' % random_port()
 
     if external:
-
         spec['oauth_redirect_uri'] = 'http://127.0.0.1:%i' % random_port()
 
-    io_loop = app.io_loop
+    event_loop = asyncio.get_running_loop()
 
     with mock.patch.object(
         jupyterhub.services.service, '_ServiceSpawner', MockServiceSpawner
@@ -310,11 +355,11 @@ def _mockservice(request, app, external=False, url=False):
             await service.start()
 
         if not external:
-            io_loop.run_sync(start)
+            await start()
 
         def cleanup():
             if not external:
-                asyncio.get_event_loop().run_until_complete(service.stop())
+                event_loop.run_until_complete(service.stop())
             app.services[:] = []
             app._service_map.clear()
 
@@ -324,26 +369,37 @@ def _mockservice(request, app, external=False, url=False):
             with raises(TimeoutExpired):
                 service.proc.wait(1)
         if url:
-            io_loop.run_sync(partial(service.server.wait_up, http=True))
+            await service.server.wait_up(http=True)
     return service
 
 
+_service_name_counter = 0
+
+
 @fixture
-def mockservice(request, app):
+def service_name():
+    global _service_name_counter
+    _service_name_counter += 1
+    name = f'test-service-{_service_name_counter}'
+    return name
+
+
+@fixture
+async def mockservice(request, app, service_name):
     """Mock a service with no external service url"""
-    yield _mockservice(request, app, url=False)
+    yield await _mockservice(request, app, name=service_name, url=False)
 
 
 @fixture
-def mockservice_external(request, app):
+async def mockservice_external(request, app, service_name):
     """Mock an externally managed service (don't start anything)"""
-    yield _mockservice(request, app, external=True, url=False)
+    yield await _mockservice(request, app, name=service_name, external=True, url=False)
 
 
 @fixture
-def mockservice_url(request, app):
+async def mockservice_url(request, app, service_name):
     """Mock a service with its own url to test external services"""
-    yield _mockservice(request, app, url=True)
+    yield await _mockservice(request, app, name=service_name, url=True)
 
 
 @fixture
@@ -369,6 +425,15 @@ def slow_spawn(app):
         yield
 
 
+@fixture
+def full_spawn(app):
+    """Fixture enabling full instrumented server via InstrumentedSpawner"""
+    with mock.patch.dict(
+        app.tornado_settings, {'spawner_class': mocking.InstrumentedSpawner}
+    ):
+        yield
+
+
 @fixture
 def never_spawn(app):
     """Fixture enabling NeverSpawner"""
@@ -437,8 +502,6 @@ def create_user_with_scopes(app, create_temp_role):
         return app.users[orm_user.id]
 
     yield temp_user_creator
-    for user in temp_users:
-        app.users.delete(user)
 
 
 @fixture
@@ -474,3 +537,66 @@ def preserve_scopes():
     scope_definitions = copy.deepcopy(scopes.scope_definitions)
     yield scope_definitions
     scopes.scope_definitions = scope_definitions
+
+
+# collect db query counts and report the top N tests by db query count
+@fixture(autouse=True)
+def count_db_executions(request, record_property):
+    if 'app' in request.fixturenames:
+        app = request.getfixturevalue("app")
+        initial_count = app.db_query_count
+        yield
+        # populate property, collected later in pytest_terminal_summary
+        record_property("db_executions", app.db_query_count - initial_count)
+    elif 'db' in request.fixturenames:
+        # some use the 'db' fixture directly for one-off database tests
+        count = 0
+        engine = request.getfixturevalue("db").get_bind()
+
+        @event.listens_for(engine, "before_execute")
+        def before_execute(conn, clauseelement, multiparams, params, execution_options):
+            nonlocal count
+            count += 1
+
+        yield
+        record_property("db_executions", count)
+    else:
+        # nothing to do, still have to yield
+        yield
+
+
+def pytest_terminal_summary(terminalreporter, exitstatus, config):
+    # collect db_executions property
+    # populated by the count_db_executions fixture
+    db_counts = {}
+    for report in terminalreporter.getreports(""):
+        properties = dict(report.user_properties)
+        db_executions = properties.get("db_executions", 0)
+        if db_executions:
+            db_counts[report.nodeid] = db_executions
+
+    total_queries = sum(db_counts.values())
+    if total_queries == 0:
+        # nothing to report (e.g. test subset)
+        return
+    n = min(10, len(db_counts))
+    terminalreporter.section(f"top {n} database queries")
+    terminalreporter.line(f"{total_queries:<6} (total)")
+    for nodeid in sorted(db_counts, key=db_counts.get, reverse=True)[:n]:
+        queries = db_counts[nodeid]
+        if queries:
+            terminalreporter.line(f"{queries:<6} {nodeid}")
+
+
+@fixture
+def service_data(service_name):
+    """Data used to create service at runtime"""
+    return {
+        "name": service_name,
+        "oauth_client_id": f"service-{service_name}",
+        "api_token": f"api_token-{service_name}",
+        "oauth_redirect_uri": "http://127.0.0.1:5555/oauth_callback-from-api",
+        "oauth_no_confirm": True,
+        "oauth_client_allowed_scopes": ["inherit"],
+        "info": {'foo': 'bar'},
+    }
