@@ -1,4 +1,6 @@
 """Test the JupyterHub entry point"""
+
+import asyncio
 import binascii
 import json
 import logging
@@ -219,7 +221,7 @@ def test_cookie_secret_env(tmpdir, request):
     assert not os.path.exists(hub.cookie_secret_file)
 
 
-def test_cookie_secret_string_():
+def test_cookie_secret_string():
     cfg = Config()
 
     cfg.JupyterHub.cookie_secret = "not hex"
@@ -233,8 +235,14 @@ def test_cookie_secret_string_():
 
 async def test_load_groups(tmpdir, request):
     to_load = {
-        'blue': ['cyclops', 'rogue', 'wolverine'],
-        'gold': ['storm', 'jean-grey', 'colossus'],
+        'blue': {
+            'users': ['cyclops', 'rogue', 'wolverine'],
+        },
+        'gold': {
+            'users': ['storm', 'jean-grey', 'colossus'],
+            'properties': {'setting3': 'three', 'setting4': 'four'},
+        },
+        'deprecated_list': ['jubilee', 'magik'],
     }
     kwargs = {'load_groups': to_load}
     ssl_enabled = getattr(request.module, "ssl_enabled", False)
@@ -242,30 +250,62 @@ async def test_load_groups(tmpdir, request):
         kwargs['internal_certs_location'] = str(tmpdir)
     hub = MockHub(**kwargs)
     hub.init_db()
+    db = hub.db
     await hub.init_role_creation()
     await hub.init_users()
     await hub.init_groups()
-    db = hub.db
+
     blue = orm.Group.find(db, name='blue')
     assert blue is not None
-    assert sorted(u.name for u in blue.users) == sorted(to_load['blue'])
+    assert sorted(u.name for u in blue.users) == sorted(to_load['blue']['users'])
+    assert blue.properties == {}
     gold = orm.Group.find(db, name='gold')
     assert gold is not None
-    assert sorted(u.name for u in gold.users) == sorted(to_load['gold'])
+    assert sorted(u.name for u in gold.users) == sorted(to_load['gold']['users'])
+    assert gold.properties == to_load['gold']['properties']
+    deprecated_list = orm.Group.find(db, name='deprecated_list')
+    assert deprecated_list is not None
+    assert deprecated_list.properties == {}
+    assert sorted(u.name for u in deprecated_list.users) == sorted(
+        to_load['deprecated_list']
+    )
 
 
-async def test_resume_spawners(tmpdir, request):
-    if not os.getenv('JUPYTERHUB_TEST_DB_URL'):
-        p = patch.dict(
-            os.environ,
-            {
-                'JUPYTERHUB_TEST_DB_URL': 'sqlite:///%s'
-                % tmpdir.join('jupyterhub.sqlite')
-            },
-        )
-        p.start()
-        request.addfinalizer(p.stop)
+@pytest.fixture
+def persist_db(tmpdir):
+    """ensure db will persist (overrides default sqlite://:memory:)"""
+    if os.getenv('JUPYTERHUB_TEST_DB_URL'):
+        # already using a db, no need
+        yield
+        return
+    with patch.dict(
+        os.environ,
+        {'JUPYTERHUB_TEST_DB_URL': f"sqlite:///{tmpdir.join('jupyterhub.sqlite')}"},
+    ):
+        yield
 
+
+@pytest.fixture
+def new_hub(request, tmpdir, persist_db):
+    """Fixture to launch a new hub for testing"""
+
+    async def new_hub():
+        kwargs = {}
+        ssl_enabled = getattr(request.module, "ssl_enabled", False)
+        if ssl_enabled:
+            kwargs['internal_certs_location'] = str(tmpdir)
+        app = MockHub(test_clean_db=False, **kwargs)
+        app.config.ConfigurableHTTPProxy.should_start = False
+        app.config.ConfigurableHTTPProxy.auth_token = 'unused'
+        request.addfinalizer(app.stop)
+        await app.initialize([])
+
+        return app
+
+    return new_hub
+
+
+async def test_resume_spawners(tmpdir, request, new_hub):
     async def new_hub():
         kwargs = {}
         ssl_enabled = getattr(request.module, "ssl_enabled", False)
@@ -388,3 +428,112 @@ def test_hub_routespec(
         assert "may not receive" in caplog.text
     else:
         assert "may not receive" not in caplog.text
+
+
+@pytest.mark.parametrize(
+    "argv, sys_argv",
+    [
+        (None, ["jupyterhub", "--debug", "--port=1234"]),
+        (["--log-level=INFO"], ["jupyterhub"]),
+    ],
+)
+def test_launch_instance(request, argv, sys_argv):
+    class DummyHub(JupyterHub):
+        def launch_instance_async(self, argv):
+            # short-circuit initialize
+            # by indicating we are going to generate config in start
+            self.generate_config = True
+            return super().launch_instance_async(argv)
+
+        async def start(self):
+            asyncio.get_running_loop().stop()
+
+    DummyHub.clear_instance()
+    request.addfinalizer(DummyHub.clear_instance)
+
+    with patch.object(sys, "argv", sys_argv):
+        DummyHub.launch_instance(argv)
+    hub = DummyHub.instance()
+    if argv is None:
+        assert hub.argv == sys_argv[1:]
+    else:
+        assert hub.argv == argv
+
+
+async def test_user_creation(tmpdir, request):
+    allowed_users = {"in-allowed", "in-group-in-allowed", "in-role-in-allowed"}
+    groups = {
+        "group": {
+            "users": ["in-group", "in-group-in-allowed"],
+        }
+    }
+    roles = [
+        {
+            "name": "therole",
+            "users": ["in-role", "in-role-in-allowed"],
+        }
+    ]
+
+    cfg = Config()
+    cfg.Authenticator.allow_all = False
+    cfg.Authenticator.allowed_users = allowed_users
+    cfg.JupyterHub.load_groups = groups
+    cfg.JupyterHub.load_roles = roles
+    ssl_enabled = getattr(request.module, "ssl_enabled", False)
+    kwargs = dict(config=cfg)
+    if ssl_enabled:
+        kwargs['internal_certs_location'] = str(tmpdir)
+    hub = MockHub(**kwargs)
+    hub.init_db()
+
+    await hub.init_role_creation()
+    await hub.init_role_assignment()
+    await hub.init_users()
+    await hub.init_groups()
+    assert hub.authenticator.allowed_users == {
+        "admin",  # added by default config
+        "in-allowed",
+        "in-group-in-allowed",
+        "in-role-in-allowed",
+        "in-group",
+        "in-role",
+    }
+
+
+async def test_recreate_service_from_database(
+    request, new_hub, service_name, service_data
+):
+    # create a hub and add a service (not from config)
+    app = await new_hub()
+    app.service_from_spec(service_data, from_config=False)
+    app.stop()
+
+    # new hub, should load service from db
+    app = await new_hub()
+    assert service_name in app._service_map
+
+    # verify keys
+    service = app._service_map[service_name]
+    for key, value in service_data.items():
+        if key in {'api_token'}:
+            # skip some keys
+            continue
+        assert getattr(service, key) == value
+
+    assert (
+        service_data['oauth_client_id'] in app.tornado_settings['oauth_no_confirm_list']
+    )
+    oauth_client = (
+        app.db.query(orm.OAuthClient)
+        .filter_by(identifier=service_data['oauth_client_id'])
+        .first()
+    )
+    assert oauth_client.redirect_uri == service_data['oauth_redirect_uri']
+
+    # delete service from db, start one more
+    app.db.delete(service.orm)
+    app.db.commit()
+
+    # start one more, service should be gone
+    app = await new_hub()
+    assert service_name not in app._service_map
