@@ -1,19 +1,17 @@
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 import json
-import string
 import warnings
 from collections import defaultdict
-from datetime import datetime, timedelta
-from functools import lru_cache
-from urllib.parse import quote, urlparse
+from datetime import timedelta
+from urllib.parse import quote, urlparse, urlunparse
 
 from sqlalchemy import inspect
 from tornado import gen, web
 from tornado.httputil import urlencode
 from tornado.log import app_log
 
-from . import orm
+from . import orm, roles, scopes
 from ._version import __version__, _check_version
 from .crypto import CryptKeeper, EncryptionUnavailable, InvalidToken, decrypt, encrypt
 from .metrics import RUNNING_SERVERS, TOTAL_USERS
@@ -21,10 +19,13 @@ from .objects import Server
 from .spawner import LocalProcessSpawner
 from .utils import (
     AnyTimeoutError,
+    _strict_dns_safe,
     make_ssl_context,
     maybe_future,
+    subdomain_hook_legacy,
     url_escape_path,
     url_path_join,
+    utcnow,
 )
 
 # detailed messages about the most common failure-to-start errors,
@@ -55,42 +56,6 @@ Common causes of this timeout, and debugging tips:
    to a number of seconds that is enough for servers to become responsive.
 """
 
-# set of chars that are safe in dns labels
-# (allow '.' because we don't mind multiple levels of subdomains)
-_dns_safe = set(string.ascii_letters + string.digits + '-.')
-# don't escape % because it's the escape char and we handle it separately
-_dns_needs_replace = _dns_safe | {"%"}
-
-
-@lru_cache()
-def _dns_quote(name):
-    """Escape a name for use in a dns label
-
-    this is _NOT_ fully domain-safe, but works often enough for realistic usernames.
-    Fully safe would be full IDNA encoding,
-    PLUS escaping non-IDNA-legal ascii,
-    PLUS some encoding of boundary conditions
-    """
-    # escape name for subdomain label
-    label = quote(name, safe="").lower()
-    # some characters are not handled by quote,
-    # because they are legal in URLs but not domains,
-    # specifically _ and ~ (starting in 3.7).
-    # Escape these in the same way (%{hex_codepoint}).
-    unique_chars = set(label)
-    for c in unique_chars:
-        if c not in _dns_needs_replace:
-            label = label.replace(c, f"%{ord(c):x}")
-
-    # underscore is our escape char -
-    # it's not officially legal in hostnames,
-    # but is valid in _domain_ names (?),
-    # and always works in practice.
-    # FIXME: We should consider switching to proper IDNA encoding
-    # for 3.0.
-    label = label.replace("%", "_")
-    return label
-
 
 class UserDict(dict):
     """Like defaultdict, but for users
@@ -160,7 +125,7 @@ class UserDict(dict):
         elif isinstance(key, str):
             orm_user = self.db.query(orm.User).filter(orm.User.name == key).first()
             if orm_user is None:
-                raise KeyError("No such user: %s" % key)
+                raise KeyError(f"No such user: {key}")
             else:
                 key = orm_user.id
         if isinstance(key, orm.User):
@@ -177,7 +142,7 @@ class UserDict(dict):
             if id not in self:
                 orm_user = self.db.query(orm.User).filter(orm.User.id == id).first()
                 if orm_user is None:
-                    raise KeyError("No such user: %s" % id)
+                    raise KeyError(f"No such user: {id}")
                 user = self.add(orm_user)
             else:
                 user = super().__getitem__(id)
@@ -223,7 +188,7 @@ class UserDict(dict):
 
         Returns dict with counts of active/pending/ready servers
         """
-        counts = defaultdict(lambda: 0)
+        counts = defaultdict(int)
         for user in self.values():
             for spawner in user.spawners.values():
                 pending = spawner.pending
@@ -336,6 +301,106 @@ class User:
             self.orm_user.groups = []
         self.db.commit()
 
+    def sync_roles(self, auth_roles):
+        """Synchronize roles with database"""
+        auth_roles_by_name = {role['name']: role for role in auth_roles}
+
+        current_user_roles = {r.name for r in self.orm_user.roles}
+        new_user_roles = set(auth_roles_by_name.keys())
+
+        granted_roles = new_user_roles.difference(current_user_roles)
+        stripped_roles = current_user_roles.difference(new_user_roles)
+
+        if granted_roles:
+            self.log.info(f"Granting user {self.name} roles(s): {granted_roles}")
+        if stripped_roles:
+            self.log.info(f"Stripping user {self.name} roles(s): {stripped_roles}")
+
+        existing_granted_roles = {
+            r.name
+            for r in self.db.query(orm.Role).filter(orm.Role.name.in_(granted_roles))
+        }
+        created_roles = existing_granted_roles.difference(granted_roles)
+
+        if created_roles:
+            self.log.info(f"Creating new roles {created_roles} in the database")
+
+        for role_name in new_user_roles:
+            if role_name in created_roles:
+                self.log.info(f"Creating new role {role_name}")
+            else:
+                self.log.debug(f"Updating existing role {role_name}")
+
+            role = auth_roles_by_name[role_name]
+            role['managed_by_auth'] = True
+
+            # creates role, or if it exists, update its `description` and `scopes`
+            try:
+                orm_role = roles.create_role(
+                    self.db, role, commit=False, reset_to_defaults=False
+                )
+            except (
+                roles.RoleValueError,
+                roles.InvalidNameError,
+                scopes.ScopeNotFound,
+            ) as e:
+                raise web.HTTPError(409, str(e))
+
+            # Update the groups, services and users for the role
+            entity_map = {
+                'groups': orm.Group,
+                'services': orm.Service,
+                'users': orm.User,
+            }
+            for key, Class in entity_map.items():
+                if key in role.keys():
+                    entities = []
+                    not_found_entities = []
+                    for entity_name in role[key]:
+                        entity = Class.find(self.db, entity_name)
+                        if entity is None:
+                            not_found_entities.append(entity_name)
+                        else:
+                            entities.append(entity)
+                    setattr(orm_role, key, entities)
+                    if not_found_entities:
+                        self.log.warning(
+                            f'Could not assign the role {role_name} to {key}:'
+                            f' {not_found_entities} not found in the database.'
+                        )
+
+        # assign the granted roles to the current user
+        for role_name in granted_roles:
+            roles.grant_role(
+                self.db,
+                entity=self.orm_user,
+                rolename=role_name,
+                commit=False,
+                managed=True,
+            )
+
+        # strip the user of roles no longer directly granted
+        for role_name in stripped_roles:
+            roles.strip_role(
+                self.db, entity=self.orm_user, rolename=role_name, commit=False
+            )
+        managed_stripped_roles = (
+            self.db.query(orm.Role)
+            .filter(
+                orm.Role.name.in_(stripped_roles) & (orm.Role.managed_by_auth == True)
+            )
+            .all()
+        )
+        for stripped_role in managed_stripped_roles:
+            if (
+                not stripped_role.users
+                and not stripped_role.services
+                and not stripped_role.groups
+            ):
+                self.db.delete(stripped_role)
+
+        self.db.commit()
+
     async def save_auth_state(self, auth_state):
         """Encrypt and store auth_state"""
         if auth_state is None:
@@ -416,9 +481,10 @@ class User:
                 yield orm_spawner
 
     def _new_orm_spawner(self, server_name):
-        """Creat the low-level orm Spawner object"""
-        orm_spawner = orm.Spawner(user=self.orm_user, name=server_name)
+        """Create the low-level orm Spawner object"""
+        orm_spawner = orm.Spawner(name=server_name)
         self.db.add(orm_spawner)
+        orm_spawner.user = self.orm_user
         self.db.commit()
         assert server_name in self.orm_spawners
         return orm_spawner
@@ -439,7 +505,7 @@ class User:
 
         # use fully quoted name for client_id because it will be used in cookie-name
         # self.escaped_name may contain @ which is legal in URLs but not cookie keys
-        client_id = 'jupyterhub-user-%s' % quote(self.name)
+        client_id = f'jupyterhub-user-{quote(self.name)}'
         if server_name:
             client_id = f'{client_id}-{quote(server_name)}'
 
@@ -460,6 +526,9 @@ class User:
             _deprecated_db_session=self.db,
             oauth_client_id=client_id,
             cookie_options=self.settings.get('cookie_options', {}),
+            cookie_host_prefix_enabled=self.settings.get(
+                "cookie_host_prefix_enabled", False
+            ),
             trusted_alt_names=trusted_alt_names,
             user_options=orm_spawner.user_options or {},
         )
@@ -472,6 +541,20 @@ class User:
             )
             spawn_kwargs.update(ssl_kwargs)
 
+        # public URLs
+        if self.settings.get("public_url"):
+            public_url = self.settings["public_url"]
+            hub = self.settings.get('hub')
+            if hub is None:
+                # only in mock tests
+                hub_path = "/hub/"
+            else:
+                hub_path = hub.base_url
+            spawn_kwargs["public_hub_url"] = urlunparse(
+                public_url._replace(path=hub_path)
+            )
+        spawn_kwargs["public_url"] = self.public_url(server_name)
+
         # update with kwargs. Mainly for testing.
         spawn_kwargs.update(kwargs)
         spawner = spawner_class(**spawn_kwargs)
@@ -558,18 +641,36 @@ class User:
     @property
     def domain(self):
         """Get the domain for my server."""
+        hook = self.settings.get("subdomain_hook", subdomain_hook_legacy)
+        return hook(self.name, self.settings['domain'], kind='user')
 
-        return _dns_quote(self.name) + '.' + self.settings['domain']
+    @property
+    def dns_safe_name(self):
+        """Get a dns-safe encoding of my name
+
+        - always safe value for a single DNS label
+        - max 40 characters, leaving room for additional components
+
+        .. versionadded:: 5.0
+        """
+        return _strict_dns_safe(self.name, max_length=40)
 
     @property
     def host(self):
         """Get the *host* for my server (proto://domain[:port])"""
-        # FIXME: escaped_name probably isn't escaped enough in general for a domain fragment
-        parsed = urlparse(self.settings['subdomain_host'])
-        h = f'{parsed.scheme}://{self.domain}'
-        if parsed.port:
-            h += ':%i' % parsed.port
-        return h
+        # if subdomains are used, use our domain
+
+        if self.settings.get('subdomain_host'):
+            parsed = urlparse(self.settings['subdomain_host'])
+            h = f"{parsed.scheme}://{self.domain}"
+            if parsed.port:
+                h = f"{h}:{parsed.port}"
+            return h
+        elif self.settings.get("public_url"):
+            # no subdomain, use public host url without path
+            return urlunparse(self.settings["public_url"]._replace(path=""))
+        else:
+            return ""
 
     @property
     def url(self):
@@ -577,8 +678,8 @@ class User:
 
         Full name.domain/path if using subdomains, otherwise just my /base/url
         """
-        if self.settings.get('subdomain_host'):
-            return f'{self.host}{self.base_url}'
+        if self.settings.get("subdomain_host"):
+            return f"{self.host}{self.base_url}"
         else:
             return self.base_url
 
@@ -587,7 +688,25 @@ class User:
         if not server_name:
             return self.url
         else:
-            return url_path_join(self.url, url_escape_path(server_name))
+            return url_path_join(self.url, url_escape_path(server_name), "/")
+
+    def public_url(self, server_name=''):
+        """Get the public URL of a server by name
+
+        Like server_url, but empty if no public URL is specified
+        """
+        # server_url will be a full URL if using subdomains
+        url = self.server_url(server_name)
+        if "://" not in url:
+            # not using subdomains, public URL may be specified
+            if self.settings.get("public_url"):
+                # add server's base_url path prefix to public host
+                url = urlunparse(self.settings["public_url"]._replace(path=url))
+            else:
+                # no public url (from subdomain or host),
+                # leave unspecified
+                url = ""
+        return url
 
     def progress_url(self, server_name=''):
         """API URL for progress endpoint for a server with a given name"""
@@ -671,14 +790,64 @@ class User:
 
         orm_server = orm.Server(base_url=base_url)
         db.add(orm_server)
-        note = "Server at %s" % base_url
-        api_token = self.new_api_token(note=note, roles=['server'])
+        note = f"Server at {base_url}"
         db.commit()
 
         spawner = self.get_spawner(server_name, replace_failed=True)
         spawner.server = server = Server(orm_server=orm_server)
         assert spawner.orm_spawner.server is orm_server
 
+        requested_scopes = spawner.server_token_scopes
+        if callable(requested_scopes):
+            requested_scopes = await maybe_future(requested_scopes(spawner))
+        if not requested_scopes:
+            # nothing requested, default to 'server' role
+            requested_scopes = orm.Role.find(db, "server").scopes
+        requested_scopes = set(requested_scopes)
+        # resolve !server filter, which won't resolve elsewhere,
+        # because this token is not owned by the server's own oauth client
+        server_filter = f"={self.name}/{server_name}"
+        requested_scopes = {
+            scope + server_filter if scope.endswith("!server") else scope
+            for scope in requested_scopes
+        }
+        # ensure activity scope is requested, since activity doesn't work without
+        activity_scope = "users:activity!user"
+        if not {activity_scope, "users:activity", "inherit"}.intersection(
+            requested_scopes
+        ):
+            self.log.warning(
+                f"Adding required scope {activity_scope} to server token, missing from Spawner.server_token_scopes. Please make sure to add it!"
+            )
+            requested_scopes |= {activity_scope}
+
+        have_scopes = roles.roles_to_scopes(roles.get_roles_for(self.orm_user))
+        have_scopes |= {"inherit"}
+        jupyterhub_client = (
+            db.query(orm.OAuthClient)
+            .filter_by(
+                identifier="jupyterhub",
+            )
+            .one()
+        )
+
+        resolved_scopes, excluded_scopes = scopes._resolve_requested_scopes(
+            requested_scopes, have_scopes, self.orm_user, jupyterhub_client, db
+        )
+        if excluded_scopes:
+            # what level should this be?
+            # for admins-get-more use case, this is going to happen for most users
+            # but for misconfiguration, folks will want to know!
+            self.log.debug(
+                "Not assigning requested scopes for %s: requested=%s, assigned=%s, excluded=%s",
+                spawner._log_name,
+                requested_scopes,
+                resolved_scopes,
+                excluded_scopes,
+            )
+
+        api_token = self.new_api_token(note=note, scopes=resolved_scopes)
+
         # pass requesting handler to the spawner
         # e.g. for processing GET params
         spawner.handler = handler
@@ -729,9 +898,9 @@ class User:
             await spawner.run_auth_state_hook(auth_state)
 
             # update spawner start time, and activity for both spawner and user
-            self.last_activity = (
-                spawner.orm_spawner.started
-            ) = spawner.orm_spawner.last_activity = datetime.utcnow()
+            self.last_activity = spawner.orm_spawner.started = (
+                spawner.orm_spawner.last_activity
+            ) = utcnow(with_tz=False)
             db.commit()
             # wait for spawner.start to return
             # run optional preparation work to bootstrap the notebook
@@ -793,7 +962,7 @@ class User:
                         )
                         self.db.delete(found)
                         self.db.commit()
-                        raise ValueError("Invalid token for %s!" % self.name)
+                        raise ValueError(f"Invalid token for {self.name}!")
                 else:
                     # Spawner.api_token has changed, but isn't in the db.
                     # What happened? Maybe something unclean in a resumed container.
@@ -806,7 +975,8 @@ class User:
                     self.new_api_token(
                         spawner.api_token,
                         generated=False,
-                        note="retrieved from spawner %s" % server_name,
+                        note=f"retrieved from spawner {server_name}",
+                        scopes=resolved_scopes,
                     )
                 # update OAuth client secret with updated API token
                 if oauth_provider:
@@ -830,9 +1000,7 @@ class User:
                 self.settings['statsd'].incr('spawner.failure.timeout')
             else:
                 self.log.exception(
-                    "Unhandled error starting {user}'s server: {error}".format(
-                        user=self.name, error=e
-                    )
+                    f"Unhandled error starting {self.name}'s server: {e}"
                 )
                 self.settings['statsd'].incr('spawner.failure.error')
                 e.reason = 'error'
@@ -840,9 +1008,7 @@ class User:
                 await self.stop(spawner.name)
             except Exception:
                 self.log.exception(
-                    "Failed to cleanup {user}'s server that failed to start".format(
-                        user=self.name
-                    ),
+                    f"Failed to cleanup {self.name}'s server that failed to start",
                     exc_info=True,
                 )
             # raise original exception
@@ -874,7 +1040,10 @@ class User:
         ssl_context = make_ssl_context(key, cert, cafile=ca)
         try:
             resp = await server.wait_up(
-                http=True, timeout=spawner.http_timeout, ssl_context=ssl_context
+                http=True,
+                timeout=spawner.http_timeout,
+                ssl_context=ssl_context,
+                extra_path="api",
             )
         except Exception as e:
             if isinstance(e, AnyTimeoutError):
@@ -888,18 +1057,14 @@ class User:
             else:
                 e.reason = 'error'
                 self.log.exception(
-                    "Unhandled error waiting for {user}'s server to show up at {url}: {error}".format(
-                        user=self.name, url=server.url, error=e
-                    )
+                    f"Unhandled error waiting for {self.name}'s server to show up at {server.url}: {e}"
                 )
                 self.settings['statsd'].incr('spawner.failure.http_error')
             try:
                 await self.stop(spawner.name)
             except Exception:
                 self.log.exception(
-                    "Failed to cleanup {user}'s server that failed to start".format(
-                        user=self.name
-                    ),
+                    f"Failed to cleanup {self.name}'s server that failed to start",
                     exc_info=True,
                 )
             # raise original TimeoutError
@@ -934,7 +1099,9 @@ class User:
             status = await spawner.poll()
             if status is None:
                 await spawner.stop()
-            self.last_activity = spawner.orm_spawner.last_activity = datetime.utcnow()
+            self.last_activity = spawner.orm_spawner.last_activity = utcnow(
+                with_tz=False
+            )
             # remove server entry from db
             spawner.server = None
             if not spawner.will_resume:
@@ -959,7 +1126,7 @@ class User:
             # trigger post-stop hook
             try:
                 await maybe_future(spawner.run_post_stop_hook())
-            except:
+            except Exception:
                 self.log.exception("Error in Spawner.post_stop_hook for %s", self)
             spawner.clear_state()
             spawner.orm_spawner.state = spawner.get_state()
